@@ -8,28 +8,33 @@
 
 #include "notifybypopup.h"
 
+#include "debug_p.h"
 #include "imageconverter.h"
 #include "knotification.h"
-#include "debug_p.h"
+#include "knotificationreplyaction.h"
 
 #include <QBuffer>
-#include <QGuiApplication>
 #include <QDBusConnection>
+#include <QGuiApplication>
 #include <QHash>
-#include <QPointer>
 #include <QMutableListIterator>
+#include <QPointer>
 #include <QUrl>
 
 #include <KConfigGroup>
 
 NotifyByPopup::NotifyByPopup(QObject *parent)
-  : KNotificationPlugin(parent)
-  , m_dbusInterface(QStringLiteral("org.freedesktop.Notifications"), QStringLiteral("/org/freedesktop/Notifications"), QDBusConnection::sessionBus())
+    : KNotificationPlugin(parent)
+    , m_dbusInterface(QStringLiteral("org.freedesktop.Notifications"), QStringLiteral("/org/freedesktop/Notifications"), QDBusConnection::sessionBus())
 {
     m_dbusServiceCapCacheDirty = true;
 
     connect(&m_dbusInterface, &org::freedesktop::Notifications::ActionInvoked, this, &NotifyByPopup::onNotificationActionInvoked);
 
+    // TODO can we check if this actually worked?
+    // probably not as this just does a DBus filter which will work but the signal might still get caught in apparmor :/
+    connect(&m_dbusInterface, &org::freedesktop::Notifications::NotificationReplied, this, &NotifyByPopup::onNotificationReplied);
+
     connect(&m_dbusInterface, &org::freedesktop::Notifications::NotificationClosed, this, &NotifyByPopup::onNotificationClosed);
 }
 
@@ -48,7 +53,7 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
         queryPopupServerCapabilities();
     } else {
         if (!sendNotificationToServer(notification, notifyConfig)) {
-            finish(notification); //an error occurred.
+            finish(notification); // an error occurred.
         }
     }
 }
@@ -65,7 +70,7 @@ void NotifyByPopup::update(KNotification *notification, const KNotifyConfig &not
 
 void NotifyByPopup::close(KNotification *notification)
 {
-    QMutableListIterator<QPair<KNotification*, KNotifyConfig> > iter(m_notificationQueue);
+    QMutableListIterator<QPair<KNotification *, KNotifyConfig>> iter(m_notificationQueue);
     while (iter.hasNext()) {
         auto &item = iter.next();
         if (item.first == notification) {
@@ -94,6 +99,8 @@ void NotifyByPopup::onNotificationActionInvoked(uint notificationId, const QStri
     if (n) {
         if (actionKey == QLatin1String("default") && !n->defaultAction().isEmpty()) {
             Q_EMIT actionInvoked(n->id(), 0);
+        } else if (actionKey == QLatin1String("inline-reply") && n->replyAction()) {
+            Q_EMIT replied(n->id(), QString());
         } else {
             bool ok;
             const int actionIndex = actionKey.toInt(&ok);
@@ -131,6 +138,23 @@ void NotifyByPopup::onNotificationClosed(uint dbus_id, uint reason)
     }
 }
 
+void NotifyByPopup::onNotificationReplied(uint notificationId, const QString &text)
+{
+    auto iter = m_notifications.find(notificationId);
+    if (iter == m_notifications.end()) {
+        return;
+    }
+
+    KNotification *n = *iter;
+    if (n) {
+        if (n->replyAction()) {
+            Q_EMIT replied(n->id(), text);
+        }
+    } else {
+        m_notifications.erase(iter);
+    }
+}
+
 void NotifyByPopup::getAppCaptionAndIconName(const KNotifyConfig &notifyConfig, QString *appCaption, QString *iconName)
 {
     KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QStringLiteral("Global"));
@@ -160,7 +184,7 @@ bool NotifyByPopup::sendNotificationToServer(KNotification *notification, const
     QString iconName;
     getAppCaptionAndIconName(notifyConfig_nocheck, &appCaption, &iconName);
 
-    //did the user override the icon name?
+    // did the user override the icon name?
     if (!notification->iconName().isEmpty()) {
         iconName = notification->iconName();
     }
@@ -172,6 +196,8 @@ bool NotifyByPopup::sendNotificationToServer(KNotification *notification, const
         text = stripRichText(text);
     }
 
+    QVariantMap hintsMap;
+
     // freedesktop.org spec defines action list to be list like
     // (act_id1, action1, act_id2, action2, ...)
     //
@@ -191,9 +217,29 @@ bool NotifyByPopup::sendNotificationToServer(KNotification *notification, const
             actionList.append(QString::number(actId));
             actionList.append(actionName);
         }
+
+        if (auto *replyAction = notification->replyAction()) {
+            const bool supportsInlineReply = m_popupServerCapabilities.contains(QLatin1String("inline-reply"));
+
+            if (supportsInlineReply || replyAction->fallbackBehavior() == KNotificationReplyAction::FallbackBehavior::UseRegularAction) {
+                actionList.append(QStringLiteral("inline-reply"));
+                actionList.append(replyAction->label());
+
+                if (supportsInlineReply) {
+                    if (!replyAction->placeholderText().isEmpty()) {
+                        hintsMap.insert(QStringLiteral("x-kde-reply-placeholder-text"), replyAction->placeholderText());
+                    }
+                    if (!replyAction->submitButtonText().isEmpty()) {
+                        hintsMap.insert(QStringLiteral("x-kde-reply-submit-button-text"), replyAction->submitButtonText());
+                    }
+                    if (replyAction->submitButtonIconName().isEmpty()) {
+                        hintsMap.insert(QStringLiteral("x-kde-reply-submit-button-icon-name"), replyAction->submitButtonIconName());
+                    }
+                }
+            }
+        }
     }
 
-    QVariantMap hintsMap;
     // Add the application name to the hints.
     // According to freedesktop.org spec, the app_name is supposed to be the application's "pretty name"
     // but in some places it's handy to know the application name itself
@@ -246,7 +292,7 @@ bool NotifyByPopup::sendNotificationToServer(KNotification *notification, const
         hintsMap[it.key()] = it.value();
     }
 
-    //FIXME - reenable/fix
+    // FIXME - re-enable/fix
     // let's see if we've got an image, and store the image in the hints map
     if (!notification->pixmap().isNull()) {
         QByteArray pixmapData;
@@ -263,10 +309,10 @@ bool NotifyByPopup::sendNotificationToServer(KNotification *notification, const
 
     const QDBusPendingReply<uint> reply = m_dbusInterface.Notify(appCaption, updateId, iconName, title, text, actionList, hintsMap, timeout);
 
-    //parent is set to the notification so that no-one ever accesses a dangling pointer on the notificationObject property
+    // parent is set to the notification so that no-one ever accesses a dangling pointer on the notificationObject property
     QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(reply, notification);
 
-    QObject::connect(watcher, &QDBusPendingCallWatcher::finished, this, [this, notification](QDBusPendingCallWatcher *watcher){
+    QObject::connect(watcher, &QDBusPendingCallWatcher::finished, this, [this, notification](QDBusPendingCallWatcher *watcher) {
         watcher->deleteLater();
         QDBusPendingReply<uint> reply = *watcher;
         m_notifications.insert(reply.argumentAt<0>(), notification);
@@ -293,7 +339,7 @@ void NotifyByPopup::queryPopupServerCapabilities()
         m_dbusServiceCapCacheDirty = false;
 
         // re-run notify() on all enqueued m_notifications
-        for (const QPair<KNotification*, KNotifyConfig> &noti : qAsConst(m_notificationQueue)) {
+        for (const QPair<KNotification *, KNotifyConfig> &noti : std::as_const(m_notificationQueue)) {
             notify(noti.first, noti.second);
         }
 
